/*
 * GMap LIMBs: Labels In the Map Border for Google Maps.
 * Version 1.1-SNAPSHOT
 *   by Stephen Battey
 *
 * http://sourceforge.net/p/gmaplimbs
 * http://www.lucidviews.net/software/gmaplimbs
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; version 3 of the License.
 * 
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU General Public License along
 * with the files in this library. If not, see  http://www.gnu.org/licenses
 * Distributed under the GNU Library or Lesser General Public License (LGPL)
 */



/* Create the namespaces used by this JavaScript library. */

if (typeof lucid === "undefined")
{
	/**
	   @namespace
	 */
	lucid = {};
}

if (typeof lucid.maps === "undefined")
{
	/**
	   @namespace
	 */
	lucid.maps = {};
}

if (typeof lucid.maps.callout === "undefined")
{
	/**
	   @namespace
	 */
	lucid.maps.callout = {};
}

if (typeof lucid.maps.callout.content === "undefined")
{
	/**
	   @namespace
	 */
	lucid.maps.callout.content = {};
}

if (typeof lucid.maps.limbs === "undefined")
{
	/**
	   @namespace
	 */
	lucid.maps.limbs = {};
}

if (typeof lucid.maps.limbs.layout === "undefined")
{
	/**
	   @namespace
	 */
	lucid.maps.limbs.layout = {};
}

if (typeof lucid.maps.limbs.tooltip === "undefined")
{
	/**
	   @namespace
	 */
	lucid.maps.limbs.tooltip = {};
}

if (typeof lucid.maps.places === "undefined")
{
	/**
	   @namespace
	 */
	lucid.maps.places = {};
}

/**
 * @class
 * Displays a callout using the InfoBox utility (see http://google-maps-utility-library-v3.googlecode.com/svn/trunk/infobox/docs/examples.html)
 * and populates the content of the info box with details read from the Google PlaceResult object.
 * 
 * @constructor
 * @param {lucid.maps.callout.InfoBoxOptions} options  The settings that control the style and content of the info box.
 */
lucid.maps.callout.InfoBox = function( options )
{
	var infoWindow;
	var currentMarker;
	
	
	function init()
	{
		infoWindow = new InfoBox( options );
	}
	
	
	this.add = function( marker )
	{
		marker.infoBoxClickEventListener = google.maps.event.addListener( marker, "click", createClickHandler( marker ) );
	};
	
	this.remove = function( marker )
	{
		if (marker.infoBoxClickEventListener)
		{
			google.maps.event.removeListener( marker.infoBoxClickEventListener );
			marker.infoBoxClickEventListener = null;
		}
		
		if (marker == currentMarker)
		{
			this.close();
		}
	};
	
	this.close = function()
	{
		infoWindow.close();
		currentMarker = null;
	};
	
	function createClickHandler( marker )
	{
		return function()
		{
			infoWindow.close();
			
			currentMarker = marker;
			
			infoWindow.setContent( createInfoBoxContent( marker ) );
			infoWindow.open( marker.getMap(), marker );
		};
	}
	
	function createInfoBoxContent( marker )
	{
		// The inner content is the content for this specific marker ...
		var innerContentDiv = (options.contentCreator) ? options.contentCreator( marker ) : lucid.maps.callout.createContentTitle( marker );
		
		// ... which we wrap in the 'content' CSS to position is correctly in the info box.
		var contentDiv = jQuery( "<div class='content'></div>" );
		contentDiv.append( innerContentDiv );
		
		// We then add the box with the arrow around that content.
		var infoBoxDiv = jQuery( "<div class='custom-info-box'><div class='arrow'></div></div>" );
		infoBoxDiv.append( contentDiv );
		
		return infoBoxDiv[0];
	}
	
	
	init();
};

/**
 * @type {object}
 * @augments InfoBoxOptions (see http://google-maps-utility-library-v3.googlecode.com/svn/trunk/infobox/docs/reference.html#InfoBoxOptions)
 * @property {function} [contentCreator]  The function that creates the content (HTML) for the info box.
 *                                        If left undefined then only the title/name of the place is shown in the info box.
 */
lucid.maps.callout.InfoBoxOptions = {};

/**
 * Create HTML content for a callout that shows the name/title of a marker.
 * Use this function as a contentCreator when creating an {@link lucid.maps.callout.InfoBox} object.
 * 
 * @param {google.maps.Marker} marker  The marker that is the subject of the callout.
 */
lucid.maps.callout.content.createContentTitle = function( marker )
{
	var titleDiv = jQuery( "<div class='place-name'>" + marker.getTitle() + "</div>" );
	
	var contentDiv = jQuery( "<div></div>" );
	contentDiv.append( titleDiv );
	
	return contentDiv;
};

/**
 * Create HTML content for a callout that shows the rating and opening hours a place.
 * Use this function as a contentCreator when creating an {@link lucid.maps.callout.InfoBox} object.
 * 
 * @param {google.maps.Marker} marker  The marker that marks the location of the place.
 *                                     The marker must also have a placeReference property, which holds the uid of the place.
 */
lucid.maps.callout.content.createContentPlaceInfo = function( marker )
{
	var contentDiv = lucid.maps.callout.content.createContentTitle( marker );
	
	var initialMessage;
	if (typeof marker.placeReference !== "undefined")
	{
		initialMessage = "Loading ...";
		requestPlaceDetail();
	}
	else
	{
		initialMessage = "No place ref for marker.";
	}
	
	var infoDiv = jQuery( "<div class='place-info'>" + initialMessage + "</div>" );
	contentDiv.append( infoDiv );
	
	return contentDiv;
	
	
	function requestPlaceDetail()
	{
		var searchService = new google.maps.places.PlacesService( marker.getMap() );
		var placeDetailRequest = { "reference": marker.placeReference };
		searchService.getDetails( placeDetailRequest, processPlaceDetail );
	}
	
	function processPlaceDetail( place, status )
	{
		// TODO Handle non "OK" states.
		
		var placeDetailParser = new lucid.maps.places.PlaceDetailParser( place );
		
		var html = "";
		if (placeDetailParser.hasRating())
		{
			html += "<p>Rating: " + placeDetailParser.getRating() + "</p>";
		}
		if (placeDetailParser.hasOpeningHours())
		{
			html += "<p>" + placeDetailParser.getOpeningHoursToday() + "</p>";
		}
		
		infoDiv.html( html );
	}
};

/**
 * @class
 * Displays a label on the border of a Google Map.
 * If multiple labels are to be displayed you are recommended to use a LimbFactory.
 * 
 * @constructor
 * @param {lucid.maps.limbs.LimbOptions|google.maps.Marker} optionsOrMarker  Settings for the LIMB to be rendered or the Marker to be tracked.
 */
lucid.maps.limbs.Limb = function( optionsOrMarker )
{
	var thisRenderer = this;
	
	var limbOptions = (optionsOrMarker.constructor == (new google.maps.Marker()).constructor) ? { marker: optionsOrMarker } : optionsOrMarker;
	
	var hidden;
	var inTheBorder;
	var marker;
	var iconDiv;
	var mapViewChangeListener;
	var syncClickWithMarker;
	var layoutManager;
	var generateTooltip;
	
	
	// This initialise function is called at the end of the constructor.
	function init()
	{
		hidden = limbOptions.hidden;
		marker = limbOptions.marker;
		syncClickWithMarker = false;
		layoutManager = (typeof limbOptions.layout === "object") ? limbOptions.layout : new lucid.maps.limbs.layout.DefaultLayoutStrategy();
		generateTooltip = (typeof limbOptions.tooltipFactory === "function") ? limbOptions.tooltipFactory : lucid.maps.limbs.tooltip.standardTooltip;
		
		if (limbOptions.independent !== false)
		{
			mapViewChangeListener = google.maps.event.addListener( thisRenderer.getMap(), "bounds_changed", handleRefresh );
		}
		
		createIconDiv();
		
		thisRenderer.refresh();
	}
	
	
	/**
	 * Destroy this instance and any associated resources.
	 * This method should be called when the instance is no longer required.
	 */
	this.destroy = function()
	{
		marker = null;
		
		if (mapViewChangeListener)
		{
			google.maps.event.removeListener( mapViewChangeListener );
		}
		
		removeIconDiv();
	};
	
	function createIconDiv()
	{
		var icon = (limbOptions.icon) ? limbOptions.icon : marker.getIcon();
		
		iconDiv = jQuery( "<div></div>" );
		iconDiv.css( "background", "url( '" + icon.url + "' ) no-repeat center center" );
		iconDiv.css( "position", "absolute" );
		iconDiv.css( "width", icon.size.width );
		iconDiv.css( "height", icon.size.height );
		iconDiv.css( "zIndex", layoutManager.getMinZIndex() );
		iconDiv.hide();
		
		if (limbOptions.clickable !== false)
		{
			if (marker.getTitle() != null)
			{
				iconDiv.attr( "title", marker.getTitle() );
			}
			
			iconDiv.css( "cursor", "pointer" );
			iconDiv.click( handleClick );
		}
		
		iconDiv.appendTo( getLimbElement() );
	}
	
	function removeIconDiv()
	{
		if (iconDiv)
		{
			iconDiv.remove();
			iconDiv = null;
		}
	}
	
	/**
	 * Temporarily hide the label.
	 * This does not remove the LIMB, it just takes it off display.
	 * Call this if you hide the map element.
	 */
	this.hide = function()
	{
		hidden = true;
		
		if (iconDiv)
		{
			iconDiv.hide();
		}
	};
	
	/**
	 * Re-display the label after being hidden with a call to 'hide'.
	 */
	this.show = function()
	{
		hidden = false;
		
		if (iconDiv)
		{
			if (inTheBorder === true)
			{
				iconDiv.show();
			}
		}
	};
	
	/**
	 * @return {google.maps.Map}  The map this LIMB is associated with.
	 */
	this.getMap = function()
	{
		return marker.getMap();
	};
	
	/**
	 * @return {google.maps.Marker}  The marker this LIMB is associated with.
	 */
	this.getMarker = function()
	{
		return marker;
	};
	
	function getMapElement()
	{
		return jQuery( thisRenderer.getMap().getDiv() );
	}
	
	function getLimbElement()
	{
		return getMapElement().parent();
	}
	
	function handleClick()
	{
		google.maps.event.trigger( thisRenderer, "click", thisRenderer );
		
		if (syncClickWithMarker === true)
		{
			google.maps.event.trigger( marker, "click" );
		}
	}
	
	/**
	 * Set whether the click event on the LIMB should cause an effective click on the associated marker.
	 * 
	 * @param {boolean} synchonised  Whether the click events should be synchonised.
	 */
	this.setSyncClickWithMarker = function( synchonised )
	{
		syncClickWithMarker = synchonised;
	};
	
	function handleRefresh()
	{
		thisRenderer.refresh();
	}
	
	/**
	 * Refresh the display of the LIMB on the page.
	 */
	this.refresh = function()
	{
		var map = this.getMap();
		var mapElement = getMapElement();
		var markerLocation = marker.getPosition();

		var mapBounds = map.getBounds();
		if (   (typeof mapBounds === "undefined")        // The getBounds method returns undefined when the map is still initialising.
		    || (mapBounds.contains( markerLocation ))  )
		{
			inTheBorder = false;
			
			iconDiv.hide();
			return;
		}
		else
		{
			inTheBorder = true;
			
			if (hidden === true)
			{
				iconDiv.hide();
			}
			else
			{
				iconDiv.show();
			}
		}

		var viewCentre = map.getCenter();
		var heading = computeHeading();
		var angle = 90 - heading;
		var angleInRadians = angle * Math.PI / 180;
		var mapWidth = mapElement.width();
		var mapHeight = mapElement.height();

		var intersection = computeIntersectionOnLeftOrRightEdge();
		if (Math.abs( intersection.y ) > (mapHeight / 2))
		// The intersection hits the top/bottom edge before the left/right edge.
		{
			intersection = computeIntersectionOnTopOrBottomEdge();
		}

		// The intersection coords are relative to the centre of the map.
		// Offset this to an origin in the top-left corner.
		// Also reverse the y-axis (positive values point up the Maths plane, but point down the screen).
		var intersectionFromTopLeft = { x: (mapWidth / 2) + intersection.x,
		                                y: (mapHeight / 2) - intersection.y };
		
		// Centre the icon on that position by applying an offset.
		// TODO Use the marker's icon's offset.
		var display = {};
		display.x = Math.round( intersectionFromTopLeft.x - (iconDiv.width() / 2) );
		display.y = Math.round( intersectionFromTopLeft.y - (iconDiv.height() / 2) );
		
		// The origin of these display coords are in the map element.
		// Offset these coords against the mapElement position to position the LIMB correctly on the screen.
		var mapPosition = mapElement.position();
		var outerWidthOffset = ( mapElement.outerWidth( true ) - mapElement.innerWidth() ) / 2;
		var outerHeightOffset = ( mapElement.outerHeight( true ) - mapElement.innerHeight() ) / 2;
		display.x = mapPosition.left + outerWidthOffset + display.x;
		display.y = mapPosition.top + outerHeightOffset + display.y;
		
		var mapDetails = { "viewCentre": viewCentre,
		                   "markerLocation": markerLocation,
		                   "heading": heading };
		
		layoutManager.layout( iconDiv, display, mapDetails );
		
		if (limbOptions.clickable !== false)
		{
			var limbLocation = computeLimbLocation();
			var distance = google.maps.geometry.spherical.computeDistanceBetween( limbLocation, markerLocation );
			
			iconDiv.attr( "title", generateTooltip( marker, distance, display, mapDetails ) );
		}
		// else: the LIMB does not respond to mouse hover; a tooltip is not needed 
		
		
		function computeHeading()
		{
			var heading = google.maps.geometry.spherical.computeHeading( viewCentre, markerLocation );

			// The function above can return a negative value.
			// The computeIntersectionOnLeftOrRightEdge/computeIntersectionOnTopOrBottomEdge rely on values being in the range 0 - 360.
			// So normalise the value now.
			while (heading > 360)
				heading -= 360;
			while (heading < 0)
				heading += 360;

			return heading;
		}

		function computeIntersectionOnLeftOrRightEdge()
		{
			// The intersection will be on either the left or right edge.
			var intersectionX = (heading < 180) ? (mapWidth / 2) : -(mapWidth / 2);

			var distanceToIntersection = intersectionX / Math.cos( angleInRadians );

			var intersectionY = distanceToIntersection * Math.sin( angleInRadians );

			return { x: intersectionX, y: intersectionY };
		}

		function computeIntersectionOnTopOrBottomEdge()
		{
			// The intersection will be on either the top or bottom edge.
			var intersectionY = ((heading > 90) && (heading < 270)) ? -(mapHeight / 2) : (mapHeight / 2);

			var distanceToIntersection = intersectionY / Math.sin( angleInRadians );

			var intersectionX = distanceToIntersection * Math.cos( angleInRadians );

			return { x: intersectionX, y: intersectionY };
		}
		
		function computeLimbLocation()
		{
			// We do this by converting the intersection coords from pixels to distance.
			// This is done by computing the scaling-factor between the element dimensions and the real-world distance.
			var mapHeightDistance = mapBounds.getNorthEast().lat() - mapBounds.getSouthWest().lat();
			var mapHeightScale = mapHeightDistance / mapHeight;
			var limbLocationLat = viewCentre.lat() + (intersection.y * mapHeightScale);
			
			var mapWidthDistance = mapBounds.getNorthEast().lng() - mapBounds.getSouthWest().lng();
			var mapWidthScale = mapWidthDistance / mapWidth;
			var limbLocationLng = viewCentre.lng() + (intersection.x * mapWidthScale);
			
			return new google.maps.LatLng( limbLocationLat, limbLocationLng );
		}
	}
	
	
	init();
};

/**
 * Indicates when the LIMB is clicked.
 *
 * @event lucid.maps.limbs.Limb#click
 * @type {object}
 */

/**
 * @type {object}
 * @property {boolean} [hidden]  Whether the LIMB is initially hidden. If so, it can be made visible with a call to show().
 * @property {boolean} [independent]  Whether this LIMB is independent of a lucid.maps.limbs.LimbFactory.
 *                                    It is more efficient for a group of LIMBs to be managed by a manager, but if
 *                                    a single LIMB is being displayed then the Limb will manage itself if
 *                                    you set this property to true. The default is true.
 * @property {google.maps.Marker} marker  The marker on the map which is to be displayed in the map border
 *                                        when the marker is outside the map's viewport.
 * @property {google.maps.Icon} [icon]  Icon specification which must contain the URL and size of the icon image to display.
 * @property {boolean} [clickable]  Whether the icon in the map border is clickable. If not defined the clickable setting is taken from the marker.
 * @property {lucid.maps.limbs.layout.LayoutStrategy} [layout]  A strategy for positioning the LIMB.
 *                                                              If not defined a DefaultLayoutStrategy will be used.
 * @property {function} [tooltipFactory]  A strategy for generating the tooltip shown when the user mouses-over the LIMB.
 *                                        NOTE: A tooltip is only shown if the LIMB is clickable.
 */
lucid.maps.limbs.LimbOptions = {};
// TODO Use the google.maps.Icon to specify the icon graphic - would need to support sprite origin and scaled size. Also read the size from a loaded image if it's not defined.

/**
 * Copy lucid.maps.limbs.LimbOptions from one instance to another.
 * Only the settings defined in the 'optionsToApply' object will be copied onto the 'options' object.
 * 
 * @param {lucid.maps.limbs.LimbOptions} options  The target instance.
 * @param {lucid.maps.limbs.LimbOptions} optionsToApply  Options that take precedence and should be copied into the target object.
 */
lucid.maps.limbs.applyLimbOptions = function( options, optionsToApply )
{
	if (typeof optionsToApply.hidden !== "undefined")
		options.hidden = optionsToApply.hidden;
	
	if (typeof optionsToApply.independent !== "undefined")
		options.independent = optionsToApply.independent;
	
	if (typeof optionsToApply.marker !== "undefined")
		options.marker = optionsToApply.marker;
	
	if (typeof optionsToApply.icon !== "undefined")
		options.icon = optionsToApply.icon;
	
	if (typeof optionsToApply.clickable !== "undefined")
		options.clickable = optionsToApply.clickable;
	
	if (typeof optionsToApply.layout !== "undefined")
		options.layout = optionsToApply.layout;
};

/**
 * @class
 * A factory class for generating LIMBs.
 * This class provides a slight performance gain by refreshing multiple LIMBs within a single event handler when the map moves.
 * This also provides convenience methods for hiding/showing LIMBs as and when the map is hidden or shown.
 * 
 * @constructor
 * @param {google.maps.Map} map  The map that is displaying the LIMBs.
 * @param {lucid.maps.limbs.LimbOptions} [defaultOptions]  Default settings for each LIMB being created by this factory.
 */
lucid.maps.limbs.LimbFactory = function( map, defaultOptions )
{
	var limbs;
	var hidden;
	
	
	// This initialise function is called at the end of the constructor.
	function init()
	{
		limbs = [];
		hidden = false;
		
		google.maps.event.addListener( map, "bounds_changed", handleMapMove );
	}
	
	/* NOTE:
		Desktop browsers trigger the "bounds_changed" event while the map is being panned.
		Mobile browsers wait until the pan is complete. This makes the LIMBs 'sticky'.
		A continuous refresh (using an interval) was trialled but map.getBounds function
		always returns the bounds from when the drag started, not it's current location.
	 */
	
	
	function handleMapMove()
	{
		refresh();
	}
	
	function refresh()
	{
		//var startTimestamp = new Date();
		
		var layout = ( (typeof defaultOptions !== "undefined") && (typeof defaultOptions.layout !== "undefined") )? defaultOptions.layout : null;
		
		if (layout)
			layout.init();
		
		for (var i=0; i<limbs.length; i++)
		{
			limbs[i].refresh();
		}
		
		if (layout)
			layout.complete();
		
		//var duration = new Date().getTime() - startTimestamp.getTime();
		//console.log( "No layout took " + duration + "ms" );
	}
	
	/**
	 * Get this factory to manage a LIMB.
	 * The LIMB should have its "independent" property set to false.
	 * If you don't have an instance of a LIMB, use the create method to create and add a new LIMB object.
	 * 
	 * @param {lucid.maps.limbs.Limb} limb  The LIMB to be managed by this factory.
	 */
	this.add = function( limb )
	{
		limbs[limbs.length] = limb;
	};
	
	/**
	 * Create a new LIMB and manage it.
	 * 
	 * @param {lucid.maps.limbs.LimbOptions|google.maps.Marker} markerOrOptions  Settings for the LIMB to be created. These settings will override the defaultOptions passed into the constructor.
	 *                                                                           Or the Marker to be tracked by the LIMB.
	 * @return {lucid.maps.limbs.Limb}  The LIMB that has been created.
	 */
	this.create = function( markerOrOptions )
	{
		var limbOptions = {};
		
		// Copy across the default options.
		if (typeof defaultOptions !== "undefined")
		{
			lucid.maps.limbs.applyLimbOptions( limbOptions, defaultOptions );
		}
		
		// Override the defaults with options specific to this LIMB.
		if (markerOrOptions.constructor == (new google.maps.Marker()).constructor)
		{
			limbOptions.marker = markerOrOptions;
		}
		else
		{
			lucid.maps.limbs.applyLimbOptions( limbOptions, markerOrOptions );
		}
		
		// This class overrides two options with fixed values.
		limbOptions.hidden = hidden;
		limbOptions.independent = false;
		
		var limb = new lucid.maps.limbs.Limb( limbOptions );
		
		this.add( limb );
		
		return limb;
	};
	
	/**
	 * Temporarily hide the labels.
	 * This does not remove the LIMBs and markers, it just takes them off display.
	 * Call this if you hide the map element.
	 */
	this.hide = function()
	{
		hidden = true;
		
		for (var i=0; i<limbs.length; i++)
		{
			limbs[i].hide();
		}
	};
	
	/**
	 * Re-display the labels after being hidden with a call to 'hide'.
	 */
	this.show = function()
	{
		hidden = false;
		
		for (var i=0; i<limbs.length; i++)
		{
			limbs[i].show();
		}
	};
	
	/**
	 * Remove all LIMBs from this collection.
	 * The LIMB will still exist, it just won't be managed by this container anymore.
	 */
	this.removeAll = function()
	{
		limbs = [];
	};
	
	/**
	 * Destroy all the LIMBs in this collection.
	 * The LIMB will be removed from the page.
	 */
	this.destroyAll = function()
	{
		for (var i=0; i<limbs.length; i++)
		{
			limbs[i].destroy();
			limbs[i] = null;
		}
		limbs = [];
	};
	
	
	init();
};

/**
 * @class
 * Manages multiple lucid.maps.limbs.Limb.
 * This provides a slight performance gain by refreshing multiple LIMBs within a single map-moved event handler.
 * This class will also create the LIMB for you, and the associated google.maps.Marker instance too.
 *
 * @constructor
 * @param {lucid.maps.limbs.LimbMarkerFactoryOptions} options  The factory options.
 */
lucid.maps.limbs.LimbMarkerFactory = function( options )
{
	var markers;
	var limbs;
	var zoomedToAll;
	
	// The zoom level used when the user clicks for info on a place.
	// This is only used when the map is zoomed out to show all places.
	// If the user has zoomed in to a custom zoom level then their custom zoom is preserved in that instance.
	var selectPlaceZoomLevel;
	
	
	// This initialise function is called at the end of the constructor.
	function init()
	{
		var defaultLimbOptions = (typeof options.limbOptions !== "undefined") ? options.limbOptions : {};
		
		markers = [];
		limbs = new lucid.maps.limbs.LimbFactory( options.map, defaultLimbOptions );
		zoomedToAll = false;
		selectPlaceZoomLevel = (typeof options.zoomLevelWhenSelected === "undefined") ? 16 : options.zoomLevelWhenSelected;
		
		google.maps.event.addListener( options.map, "zoom_changed", handleZoomChanged );
	}
	
	
	function handleZoomChanged()
	{
		// The map is now at a user-chosen zoom level.
		zoomedToAll = false;
	}
	
	/**
	 * Remove all markers and LIMBs created by this manager.
	 */
	this.removeAll = function()
	{
		for (var i=0; i<markers.length; i++)
		{
			if (typeof options.infoBox !== "undefined")
				options.infoBox.remove( markers[i] );
			
			markers[i].setMap( null );
			markers[i] = null;
		}
		markers = [];
		
		limbs.destroyAll();
	};
	
	/**
	 * Add a new marker and a LIMB to the map.
	 * The map instance and label styling are set in the options passed into the manager's constructor.
	 * 
	 * @param {google.maps.LatLng} position  The location to be marked.
	 * @param {string} title  The name/title of the location. Used as the title of the marker created on the map.
	 * @param {string} [placeReference]  Optionally associated the marker with a place using the place reference.
	 */
	this.add = function( position, title, placeReference )
	{
		// TODO Take a deep copy.
		var markerOptions = options.markerOptions;
		markerOptions.position = position;
		markerOptions.title = title;
		markerOptions.map = options.map;
		
		var marker = new google.maps.Marker( markerOptions );
		if (typeof placeReference !== "undefined")
		{
			marker.placeReference = placeReference;
		}
		markers[markers.length] = marker;
		
		
		var limbOptions = {};
		limbOptions.marker = marker;
		
		var limb = limbs.create( limbOptions );
		
		
		if (typeof options.infoBox !== "undefined")
		{
			options.infoBox.add( marker );
		}
		google.maps.event.addListener( marker, "click", createClickMarkerHandler( marker ) );
		google.maps.event.addListener( limb, "click", createClickLimbHandler( marker ) );
	};

	function createClickMarkerHandler( marker )
	{
		return function()
		{
			if (typeof options.infoBox === "undefined")
			{
				options.map.setCenter( marker.getPosition() );
			}
			// else: The info box will appear when the marker is clicked.
			//       Let the map centre be centred around the info box.
			
			if (zoomedToAll === true)
			{
				if (selectPlaceZoomLevel != null)
				{
					options.map.setZoom( selectPlaceZoomLevel );
				}
				// else: the zoom is turned off
			}
			// else: keep the map at the user-chosen zoom level
		}
	}
	
	function createClickLimbHandler( marker )
	{
		return function()
		{
			// Perform the same action as if the user clicked on the marker.
			google.maps.event.trigger( marker, "click" );
		};
	}
	
	/**
	 * Zoom and pan the map so that all markers are in view.
	 * You can optionally specify additional locations to be included in the map view by setting LatLng in the options parameter.
	 *
	 * @param {object} zoomOptions
	 * @param {google.maps.LatLng[]} [zoomOptions.additionalLocations]  Locations, in addition to the markers, which should be kept in view when the map view is changed.
	 */
	this.zoomToAll = function( zoomOptions )
	{
		var locations = [];
		
		for (var i=0; i<markers.length; i++)
		{
			locations[locations.length] = markers[i].getPosition();
		}
		
		if (zoomOptions.additionalLocations)
		{
			for (var i=0; i<zoomOptions.additionalLocations.length; i++)
			{
				locations[locations.length] = zoomOptions.additionalLocations[i];
			}
		}
		
		if (locations.length > 0)
		{
			var bounds = new google.maps.LatLngBounds( locations[0], locations[0] );
			for (var i=0; i<locations.length; i++)
			{
				bounds = bounds.extend( locations[i] );
			}
			
			options.map.fitBounds( bounds );
		}
		
		zoomedToAll = true;
	};
	
	/**
	 * Temporarily hide the labels.
	 * This does not remove the LIMBs and markers, it just takes them off display.
	 * Call this if you hide the map element.
	 */
	this.hide = function()
	{
		limbs.hide();
		// NB: The markers are part of the map and should be hidden by hiding the map element.
	};
	
	/**
	 * Re-display the labels after being hidden with a call to 'hide'.
	 */
	this.show = function()
	{
		limbs.show();
		// NB: The markers are part of the map and should be shown by showing the map element.
	};
	
	/**
	 * Destroy this instance and any associated resources.
	 * This method should be called when the instance is no longer required.
	 */
	this.destroy = function()
	{
		this.removeAll();
	};
	
	/**
	 * @return {number}  The total number of places being labelled (includes any labels currently being hidden).
	 */
	this.getNumberOfMarkers = function()
	{
		return markers.length;
	};
	
	
	init();
};


/**
 * @type {object}
 * @property {google.maps.Map} map  The map the markers and LIMBs are to be associated with.
 * @property {google.maps.MarkerOptions} markerOptions  The settings for the marker shown on the map for these labels.
 *                                                      The LimbMarkerFactory will ignore the position property. This will be set when the places are added with a call to addPlaces.
 * @property {lucid.maps.limbs.LimbOptions} [limbOptions] The settings for the LIMB shown on the map for these labels.
 *                                                        The LimbMarkerFactory will ignore the marker property. This will be set to the place's marker.
 *                                                        The LimbMarkerFactory will ignore the independent property. This will always be false.
 *                                                        If undefined the LIMB will take its styling from the marker.
 * @property {lucid.maps.limbs.InfoBox} [infoBox]  The class that displays an info box when the user clicks on the label.
 *                                                 Leave undefined if no info box should appear.
 * @property {number} [zoomLevelWhenSelected]  The zoom level to be used when a marker is clicked.
 *                                             Set this to null if you do not want the zoom to change.
 *                                             Defaults to 16 if not defined.
 */
lucid.maps.limbs.LimbMarkerFactoryOptions = {};

/**
 * @class
 * A layout strategy that simply positions elements in their default position.
 * There is no logic to avoid overlapping elements or to sort in z-order.
 * You do not need to call the start and stop methods on this LayoutStrategy.
 */
lucid.maps.limbs.layout.DefaultLayoutStrategy = function()
{
	lucid.maps.limbs.layout.LayoutStrategy.apply( this, [ {} ] );
	
	
	this.layout = function( div, position, mapDetails )
	{
		div.css( "left", position.x );
		div.css( "top", position.y );
		div.css( "zIndex", this.getMinZIndex() );
	};
	
};

/**
 * @class
 * A strategy class for positioning the LIMB elements.
 * This base class does nothing. Use one of the concrete sub-classes, e.g. DefaultLayoutStrategy.
 * 
 * @constructor
 * @param {lucid.maps.limbs.layout.LayoutStrategyOptions} options  The options controlling the layout.
 */
lucid.maps.limbs.layout.LayoutStrategy = function( options )
{
	
	this.init = function()
	{
	};
	
	this.layout = function( div, position, mapDetails )
	{
	};
	
	this.complete = function()
	{
	};
	
	this.getMaxZIndex = function()
	{
		return (typeof options.maxZ === "number") ? options.maxZ : 600;
	}
	
	this.getMinZIndex = function()
	{
		return (typeof options.minZ === "number") ? options.minZ : 500;
	}
	
};


/**
 * @type {object}
 * @property {number} [maxZ]  The maximum z-index to be applied to the LIMB DOM element.
 *                            The maximum z-index is applied to features nearest to the map view.
 *                            If not defined, a default value of 600 is used.
 * @property {number} [minZ]  The minimum z-index to be applied to the LIMB DOM element.
 *                            The minimum z-index is applied to features furthest from the map view.
 *                            If not defined, a default value of 500 is used.
 */
lucid.maps.limbs.layout.LayoutStrategyOptions = {};

/**
 * @class
 * A strategy that orders the z-index of all the LIMBs according to the distance to the marker.
 * LIMBs representing markers that are closest to the map will be shown above those that are further away.
 * 
 * @constructor
 * @param {lucid.maps.limbs.layout.OrderedLayoutStrategyOptions} options  The options controlling the layout.
 */
lucid.maps.limbs.layout.OrderedLayoutStrategy = function( options )
{
	lucid.maps.limbs.layout.LayoutStrategy.apply( this, [options] );

	
	var limbs;
	var maxDistance;
	
	
	this.layout = function( div, position, mapDetails )
	{
		var distance = google.maps.geometry.spherical.computeDistanceBetween( mapDetails.viewCentre, mapDetails.markerLocation );
		
		if (distance > maxDistance)
		{
			maxDistance = distance;
		}
		
		limbs[limbs.length] = { element: div,
		                        position: position,
		                        distance: distance };
	};
	
	this.init = function()
	{
		limbs = [];
		maxDistance = 0;
	};
	
	this.complete = function()
	{
		var zIndexMin = this.getMinZIndex();
		var zIndexMax = this.getMaxZIndex();
		var zIndexRange = zIndexMax - zIndexMin;
		var zIndexDistanceRatio = zIndexRange / maxDistance;
		
		for (var i=0; i<limbs.length; i++)
		{
			var limb = limbs[i];
			
			limb.element.css( "left", limb.position.x );
			limb.element.css( "top", limb.position.y );
			limb.element.css( "zIndex", computeLimbElementZIndex( limb.distance ) );
		}
		
		limbs = null;
		maxDistance = null;
		
		function computeLimbElementZIndex( distance )
		{
			return Math.floor( zIndexMax - (zIndexDistanceRatio * distance) );
		}
	};
	
};


/**
 * @type {object}
 * @augments lucid.maps.limbs.layout.LayoutStrategyOptions
 */
lucid.maps.limbs.layout.OrderedLayoutStrategyOptions = {};

/**
 * @class
 * A strategy that orders the z-index of LIMBs that are overlapping.
 * LIMBs representing markers that are closest to the map will be shown above those that are further away.
 * 
 * @constructor
 * @param {lucid.maps.limbs.layout.ZShiftedLayoutStrategyOptions} options  The options controlling the layout.
 */
lucid.maps.limbs.layout.ZShiftedLayoutStrategy = function( options )
{
	lucid.maps.limbs.layout.LayoutStrategy.apply( this, [options] );

	
	var limbs;
	var super_ = this.createSuperReferences();
	
	
	this.layout = function( div, position, mapDetails )
	{
		super_.layout( div, position, mapDetails );
		
		limbs[limbs.length] = { element: div,
		                        position: position,
		                        mapDetails: mapDetails };
	};
	
	this.init = function()
	{
		super_.init();
		
		limbs = [];
	};
	
	this.complete = function()
	{
		var zIndexMin = this.getMinZIndex();
		var zIndexMax = this.getMaxZIndex();
		var zIndexRange = zIndexMax - zIndexMin;
		var zIndexDistanceRatio;
		
		var overlappingLimbs;
		
		for (var i=0; i<limbs.length; i++)
		{
			var limb = limbs[i];
			if (limb !== null)
			{
				// For each LIMB we're going to look for overlapping LIMBs.
				
				overlappingLimbs = [];
				
				// Remove the original LIMB from the array now.
				// Otherwise we will encounter infinite recursion when we look for LIMBs that overlap the overlapping LIMBs.
				moveLimbIntoOverlappingArray( limb, i );
				
				lookForOverlappingLimbs( limb );
				
				if (overlappingLimbs.length === 1)
				{
					// No overlapping LIMBs. The basic position is fine.
					
					limb.element.css( "left", limb.position.x );
					limb.element.css( "top", limb.position.y );
				}
				else
				{
					computeZIndexDistanceRatio();
					
					for (var j=0; j<overlappingLimbs.length; j++)
					{
						var overlappingLimb = overlappingLimbs[j];
						
						overlappingLimb.element.css( "left", overlappingLimb.position.x );
						overlappingLimb.element.css( "top", overlappingLimb.position.y );
						overlappingLimb.element.css( "zIndex", computeLimbElementZIndex( overlappingLimb.distance ) );
					}
				}
			}
		}
		
		limbs = null;
		
		super_.complete();
		
		
		function lookForOverlappingLimbs( limb )
		{
			for (var i=0; i<limbs.length; i++)
			{
				var otherLimb = limbs[i];
				if (otherLimb !== null)
				{
					if (isOverlappingLimb( limb.element, otherLimb.element ))
					{
						moveLimbIntoOverlappingArray( otherLimb, i );
						
						lookForOverlappingLimbs( otherLimb );
					}
				}
			}
		}
		
		function isOverlappingLimb( limb, otherLimb )
		{
			// http://stackoverflow.com/questions/14012766/detecting-whether-two-divs-overlap
		      var x1 = limb.offset().left;
		      var y1 = limb.offset().top;
		      var h1 = limb.outerHeight(true);
		      var w1 = limb.outerWidth(true);
		      var b1 = y1 + h1;
		      var r1 = x1 + w1;
		      var x2 = otherLimb.offset().left;
		      var y2 = otherLimb.offset().top;
		      var h2 = otherLimb.outerHeight(true);
		      var w2 = otherLimb.outerWidth(true);
		      var b2 = y2 + h2;
		      var r2 = x2 + w2;

		      if (b1 < y2 || y1 > b2 || r1 < x2 || x1 > r2) return false;
		      return true;
		}
		
		function moveLimbIntoOverlappingArray( limb, limbIndex )
		{
			limbs[limbIndex] = null;
			overlappingLimbs[overlappingLimbs.length] = limb;
		}
		
		function computeZIndexDistanceRatio()
		{
			var maxDistance = 0;
			
			for (var i=0; i<overlappingLimbs.length; i++)
			{
				var limb = overlappingLimbs[i];
				
				limb.distance = google.maps.geometry.spherical.computeDistanceBetween( limb.mapDetails.viewCentre, limb.mapDetails.markerLocation );
				
				if (limb.distance > maxDistance)
				{
					maxDistance = limb.distance;
				}
			}
			
			zIndexDistanceRatio = zIndexRange / maxDistance;
		}
		
		function computeLimbElementZIndex( distance )
		{
			return Math.floor( zIndexMax - (zIndexDistanceRatio * distance) );
		}
	};
	
};


/**
 * @type {object}
 * @augments lucid.maps.limbs.layout.LayoutStrategyOptions
 */
lucid.maps.limbs.layout.ZShiftedLayoutStrategyOptions = {};


lucid.maps.limbs.tooltip.standardTooltip = function( marker, distance, display, mapDetails )
{
	return marker.getTitle() + "\n" + lucid.maps.limbs.tooltip.distanceInVariableFormat( distance );
};


lucid.maps.limbs.tooltip.distanceInVariableFormat = function( distance )
{
	if (distance < 999.5)   // whole number of metres  (0 m - 999 m)
	{
		return lucid.maps.limbs.tooltip.distanceInWholeUnits( distance ) + " m";
	}
	else if (distance < 9950)   //  (1 km - 9.9 km)
	{
		return lucid.maps.limbs.tooltip.distanceWithOneDecimalPlace( distance / 1000 ) + " km";
	}
	else  // whole number of kilometres (10 km .... )
	{
		return lucid.maps.limbs.tooltip.distanceInWholeUnits( distance / 1000 ) + " km";
	}
};

lucid.maps.limbs.tooltip.distanceInWholeUnits = function( distance )
{
	return Math.floor( distance + 0.5 );
};

lucid.maps.limbs.tooltip.distanceWithOneDecimalPlace = function( distance )
{
	return Math.floor( (distance * 10) + 0.5 ) / 10;
};

/**
 * @class
 * Searches for places and adds labels (Marker and Limb) to a map.
 * This class uses the google.maps.places.PlacesService to find places
 * and displays them using a lucid.maps.limbs.LimbMarkerFactory.
 *
 * This logic can optionally be synchronised to a button (lucid.maps.places.LabelPlacesButton).
 *
 * @constructor
 * @param {lucid.maps.places.LabelPlacesOptions} options  Configuration for the labelling of places.
 */
lucid.maps.places.LabelPlaces = function( options )
{
	var thisLabeller = this;
	
	var markerManager;
	var searchService;
	var searchOrigin;
	var searchTypes;
	
	
	// This initialise function is called at the end of the constructor.
	function init()
	{
		if (options.button)
			google.maps.event.addListener( options.button, "click", handleButtonClick );
		
		markerManager = new lucid.maps.limbs.LimbMarkerFactory( options );
		searchService = new google.maps.places.PlacesService( options.map );
		searchTypes = options.types;
	}
	
	
	/**
	 * Set the type(s) of place this instance searches for.
	 *  
	 * @param {string[]} types  A list of Google places types.
	 */
	this.setPlaceTypes = function( types )
	{
		searchTypes = types;
	};
	
	function handleButtonClick()
	{
		if (options.button.getState() == lucid.maps.places.LabelPlacesState.ACTIVE)
		{
			thisLabeller.removePlaces();
		}
		else if (options.button.getState() == lucid.maps.places.LabelPlacesState.PENDING)
		{
			// Ignore any further clicks until the current search completes.
		}
		else
		{
			thisLabeller.addPlaces();
		}
	}
	
	/**
	 * Add places to the map.
	 * The Google PlacesService will be called to find the places to show. The locations
	 * returned are then marked on the map with a Marker and a LIMB.
	 * You can remove the labels with a call to 'removePlaces'.
	 */
	this.addPlaces = function()
	{
		if (options.button)
			options.button.setPending();
		
		searchOrigin = options.map.getCenter();
		
		var placeSearchRequest =
		{
			location: searchOrigin,
			// You can't specify a max radius when ranking by distance.
			// So we'll cap the radius when we process the results.
			//radius: options.maxDistance,
			rankBy: google.maps.places.RankBy.DISTANCE,
			types: searchTypes
		};
		searchService.nearbySearch( placeSearchRequest, handleSearchResults );
	}
	
	function handleSearchResults( results, status, pagination )
	{
		if (status == google.maps.places.PlacesServiceStatus.OK)
		{
			var processedResults = processSearchResults( results );
			if (processedResults.length > 0)
			{
				handleNonZeroSearchResults( processedResults );
			}
			else
			{
				handleZeroSearchResults();
			}
		}
		else if (status == google.maps.places.PlacesServiceStatus.ZERO_RESULTS)
		{
			handleZeroSearchResults();
		}
		else
		{
			handleErrorSearchResult();
		}
	}
	
	function handleErrorSearchResult()
	{
		if (options.button)
			options.button.setError();
	}
	
	function handleZeroSearchResults()
	{
		if (options.button)
			options.button.setNoResults();
	}
	
	function handleNonZeroSearchResults( processedResults )
	{
		if (options.button)
			options.button.setActive();
		
		google.maps.event.trigger( thisLabeller, "pre_add_labels" );
		
		for (var i=0; i<processedResults.length; i++)
		{
			var result = processedResults[i];
			markerManager.add( result.geometry.location, result.name, result.reference );
		}
		
		markerManager.zoomToAll( { "additionalLocations": [ searchOrigin ] } );
		
		google.maps.event.trigger( thisLabeller, "post_add_labels" );
	}
	
	function processSearchResults( results )
	{
		var processedResults = [];
		
		for (var i=0; i<results.length; i++)
		{
			if (typeof options.maxPlaces === "number")
			{
				// Stop when we've reached the maximum places to be shown.
				if (i >= options.maxPlaces)
				{
					break;
				}
			}
			
			if (typeof options.maxDistance === "number")
			{
				// Stop when we've reach a result too far away.
				// NB: We've ranked the results by distance so that
				//     once we find a result too far away subsequent results will always be further away.
				var distanceToResults = google.maps.geometry.spherical.computeDistanceBetween( searchOrigin, results[i].geometry.location );
				if (distanceToResults > options.maxDistance)
				{
					break;
				}
			}
			
			processedResults[processedResults.length] = results[i];
		}
		
		return processedResults;
	}
	
	/**
	 * Remove places from the map.
	 */
	this.removePlaces = function()
	{
		if (markerManager.getNumberOfMarkers() > 0)
		{
			google.maps.event.trigger( thisLabeller, "pre_remove_labels" );

			if (options.button)
				options.button.setIdle();

			markerManager.removeAll();

			google.maps.event.trigger( thisLabeller, "post_remove_labels" );
		}
	};
	
	/**
	 * Temporarily hide all places.
	 * This does not remove the LIMBs and markers, it just takes them off display.
	 * Call this if you hide the map element.
	 */
	this.hide = function()
	{
		markerManager.hide();
	};
	
	/**
	 * Re-display the places after being hidden with a call to 'hide'.
	 */
	this.show = function()
	{
		markerManager.show();
	};
	
	/**
	 * @return {number}  The total number of places being labelled (includes any labels currently being hidden).
	 */
	this.getNumberOfPlaces = function()
	{
		return markerManager.getNumberOfMarkers();
	};
	
	
	init();
};

/**
 * Raised before new places are labelled on the map.
 *
 * @event lucid.maps.places.LabelPlaces#pre_add_labels
 * @type {object}
 */

/**
 * Raised after labels have been added to the map.
 * 
 * @event lucid.maps.places.LabelPlaces#post_add_labels
 * @type {object}
 */

/**
 * Raised before removing place labels from the map.
 *
 * @event lucid.maps.places.LabelPlaces#pre_remove_labels
 * @type {object}
 */

/**
 * Raised after labels have been removed from the map.
 *
 * @event lucid.maps.places.LabelPlaces#post_remove_labels
 * @type {object}
 */

/**
 * @type {object}
 * @augments lucid.maps.limbs.LimbMarkerFactoryOptions
 * @property {string[]} types  The types of places to show.
 * @property {number} maxDistance  The maximum distance to find places.
 * @property {number} maxPlaces  The maximum number of places to show.
 * @property {lucid.maps.places.LabelPlacesButton} [button]  The button which will activate the places being shown.
 *                                                           If undefined, you must trigger the places to be shown by calling the addPlaces function.
 */
lucid.maps.places.LabelPlacesOptions = {};

/**
 * @class
 * Manages the button that controls the display of LIMBs.
 * Specifically, this class is responsible for changing the button graphic to reflect the current state.
 *
 * @constructor
 * @param {lucid.maps.places.LabelPlacesButtonOptions} options  Configuration for the button.
 */
lucid.maps.places.LabelPlacesButton = function( options )
{
	var thisButton = this;
	
	// The state controls what happens when the button is clicked
	// and is reflected by the icon image used for the button graphic.
	var state;
	
	// Some states revert to idle after a given timeout.
	var revertToIdleTimer;
	var revertToIdleTimeout = 5000;
	
	var buttonElement;
	var graphics;
	
	
	// This initialise function is called at the end of the constructor.
	function init()
	{
		buttonElement = jQuery( options.buttonDiv );
		buttonElement.click( handleClick );
		
		graphics = jQuery( '.graphics', buttonElement );
		
		thisButton.setIdle();
	}
	
	
	function handleClick()
	{
		google.maps.event.trigger( thisButton, 'click' );
	}
	
	/**
	 * Make the button visible, after a call to hide.
	 */
	this.show = function()
	{
		buttonElement.show();
	};
	
	/**
	 * Hide the button graphic.
	 */
	this.hide = function()
	{
		buttonElement.hide();
	};
	
	/**
	 * Set this button into the "idle" state.
	 * @see The lucid.maps.places.LabelPlacesState enumeration.
	 */
	this.setIdle = function()
	{
		setState( lucid.maps.places.LabelPlacesState.IDLE );
	};
	
	/**
	 * Set this button into the "pending results" state.
	 * @see The lucid.maps.places.LabelPlacesState enumeration.
	 */
	this.setPending = function()
	{
		setState( lucid.maps.places.LabelPlacesState.PENDING );
	};
	
	/**
	 * Set this button into the "active" state.
	 * @see The lucid.maps.places.LabelPlacesState enumeration.
	 */
	this.setActive = function()
	{
		setState( lucid.maps.places.LabelPlacesState.ACTIVE );
	};
	
	/**
	 * Set this button into the "error encountered" state.
	 * @see The lucid.maps.places.LabelPlacesState enumeration.
	 */
	this.setError = function()
	{
		setState( lucid.maps.places.LabelPlacesState.ERROR );
		setRevertToIdle();
	};
	
	/**
	 * Set this button into the "no results found" state.
	 * @see The lucid.maps.places.LabelPlacesState enumeration.
	 */
	this.setNoResults = function()
	{
		setState( lucid.maps.places.LabelPlacesState.NO_RESULTS );
		setRevertToIdle();
	};
	
	function setState( newState )
	{
		clearRevertToIdle();
		
		if (state)
		{
			graphics.removeClass( state );
		}
		
		state = newState;
		
		graphics.addClass( state );
	};
	
	/**
	 * @return {lucid.maps.places.LabelPlacesState}  The current state of this button.
	 */
	this.getState = function()
	{
		return state;
	};
	
	function setRevertToIdle()
	{
		clearRevertToIdle();
		revertToIdleTimer = window.setTimeout( revertToIdle, revertToIdleTimeout );
	}
	
	function clearRevertToIdle()
	{
		if (revertToIdleTimer)
		{
			window.clearTimeout( revertToIdleTimer );
			revertToIdleTimer = null;
		}
	}
	
	function revertToIdle()
	{
		thisButton.setIdle();
	}
	
	
	init();
};

/**
 * Indicates when the button is clicked.
 *
 * @event lucid.maps.places.LabelPlacesButton#click
 * @type {object}
 */

/**
 * An enumeration of states for the process-lifecycle of labelling places.
 * The {@link lucid.maps.places.LabelPlacesButton} will adjust its appearance to reflect the current state.
 * @enum
 * @default IDLE
 */
lucid.maps.places.LabelPlacesState =
{
	/** No search results are being shown on the map.
	 *  @const
	 *  @type {string} */
	IDLE: "idle",
	
	/** A search is currently being performed.
	 *  @const
	 *  @type {string} */
	PENDING: "pending",
	
	/** A search has been triggered and places are being shown on the map.
	 *  @const
	 *  @type {string} */
	ACTIVE: "active",
	
	/** A search was triggered but there was an error finding nearby places.
	 *  @const
	 *  @type {string} */
	ERROR: "error",
	
	/** A search was triggered but no results were found.
	 *  @const
	 *  @type {string} */
	NO_RESULTS: "no-results"
};

/**
 * @type {object}
 * @property {Node} buttonDiv  The DOM element used to show/hide the place labels.
 */
lucid.maps.places.LabelPlacesButtonOptions = {};

/**
 * @class
 * Parses information from the place object returned from the Google Places service.
 * This class acts as a wrapper around the Google place object
 * and provides methods to parse information from the wrapped object.
 * 
 * @constructor
 * @param {google.maps.places.PlaceResult} place  The place we want the details from.
 */
lucid.maps.places.PlaceDetailParser = function( place )
{
	
	/**
	 * Check if the place has been given a rating.
	 * 
	 * @return {boolean}  True if the place has a rating.
	 */
	this.hasRating = function()
	{
		return (typeof place.rating === "number");
	};
	
	/**
	 * Get the rating given to the place.
	 * This may be undefined. Use the hasRating function to test if the rating exists.
	 * 
	 * @return {number}  The place's rating.
	 */
	this.getRating = function()
	{
		return place.rating;
	};
	
	/**
	 * Check if the place has defined its opening hours.
	 * 
	 * @return {boolean}  True if the place has opening hours defined.
	 */
	this.hasOpeningHours = function()
	{
		return (typeof place.opening_hours !== "undefined") && (typeof place.opening_hours.periods !== "undefined") && (typeof place.opening_hours.periods.length === "number");
	};
	
	/**
	 * Get the opening hours for the current day of the week.
	 * This may be undefined. Use the hasOpeningHours function to test if the opening hours exist.
	 * 
	 * @return {string}  The opening hours for the current day of the week.
	 */
	this.getOpeningHoursToday = function()
	{
		return this.getOpeningHours( new Date().getDay() );
	};
	
	/**
	 * Get the opening hours for the specified day of the week.
	 * This may be undefined. Use the hasOpeningHours function to test if the opening hours exist.
	 * 
	 * @param {number} day  The zero-based index of the day of the week. With 0 being Monday.
	 * @return {string}  The opening hours for the current day of the week.
	 */
	this.getOpeningHours = function( day )
	{
		var periods = [];
		for (var i=0; i<place.opening_hours.periods.length; i++)
		{
			if (place.opening_hours.periods[i].open.day == day)
			{
				periods[periods.length] = place.opening_hours.periods[i];
			}
		}
		
		var openingHoursText;
		if (periods.length > 0)
		{
			openingHoursText = "Open ";
			for (var i=0; i<periods.length; i++)
			{
				if (i > 0)
				{
					openingHoursText += ", ";
				}
				openingHoursText += periods[i].open.time + " - " + periods[i].close.time;
			}
		}
		else
		{
			openingHoursText = "Closed";
		}
		
		return openingHoursText;
	};
	
	// TODO Parse "reviews[n].text"
};
